/********************************************************************
*                                                                   *
*    C adaptation written by Angela Feistel, 29 Dec 2003            *
*                                                                   *
*    ANSI C demo program to compute thermodynamic properties        *
*    of seawater from a thermodynamic potential function g(S,t,p).  *
*    Details are published in "Progress in Oceanography",           *
*    58 (2003) 43-114, entitled                                     *
*                                                                   *
*    "A new extended Gibbs thermodynamic potential of seawater"     *
*                                                                   *
*               by R.Feistel                                        *
*                                                                   *
*                                                                   *
*                 Dr. Rainer Feistel                                *
*                 Institut fuer Ostseeforschung, IOW                *
*                 D-18119 Warnemuende                               *
*                 Seestrasse 15                                     *
*                 GERMANY                                           *
*                                                                   *
*                 Tel: +49-381-5197-152                             *
*                 Fax: +49-381-5197-4818                            *
*                 Email: rainer.feistel@io-warnemuende.de           *
*                                                                   *
********************************************************************/



#include <stdio.h>
#include <math.h>


#define psu	35.0
#define degC	20.0
#define dbar	2000.0


// Specify a COMMON block for array of polynomial coefficients gc(8,8,7)
#define maxS 7
#define maxT 7
#define maxP 6


// array of coeffs COEFFS03
double gc[maxS+1][maxT+1][maxP+1]; 
// init coeffs
void gc_init(void);	


// functions
double Gxyz(int nx, int ny, int nz, double x, double y, double z) ;
double GSTP03(int nS, int nT, int nP, double S, double tabs, double pabs);
double PotTemp(double Spsu, double tdegC, double pdBar, double prefdBar);


// main program
void main() {

	double S, T, P;
	gc_init();

	
	S = psu;
	T = degC + 273.15;          //degC -> K
	P = dbar * 10000 + 101325;	//dBar -> Pa


	printf( "S = %.2f psu \n",psu);
	printf( "T = %.2f dC \n",degC);
	printf( "P = %.2f dbar \n",dbar);
	printf( " \n");
	printf( "free enthalpy : %.12e J/kg \n",  GSTP03(0, 0, 0, S, T, P));
	printf( "chem. pot.    : %.12e J/kg psu \n", GSTP03(1, 0, 0, S, T, P)); 
	printf( "entropy       : %.12e J/kgK \n", -GSTP03(0, 1, 0, S, T, P)); 
	printf( "density       : %.12e kg/m^3 \n", 1 / GSTP03(0, 0, 1, S, T, P)); 
	printf( "heat capacity : %.12e J/kgK \n", -T * GSTP03(0, 2, 0, S, T, P));
	printf( "Ch. pot. H2O  : %.12e J/kg \n", GSTP03(0, 0, 0, S, T, P) - S * GSTP03(1, 0, 0, S, T, P));
	printf( "therm. exp.   : %.12e 1/K \n", GSTP03(0, 1, 1, S, T, P) / GSTP03(0, 0, 1, S, T, P));
	printf( "compressib.   : %.12e 1/Pa \n", -GSTP03(0, 0, 2, S, T, P) / GSTP03(0, 0, 1, S, T, P));
	printf( "lapse rate    : %.12e K/Pa \n", -GSTP03(0, 1, 1, S, T, P) / GSTP03(0, 2, 0, S, T, P));
	printf( "pot. temp.    : %.12e degC \n", PotTemp(S, degC, dbar, 0));


	/* Printout:
	S = 35.00 psu
	T = 20.00 dC
	P = 2000.00 dbar
	free enthalpy : 1.658318067148e+004 J/kg
	chem. pot.    : 6.000993666928e+001 J/kg psu
	entropy       : 2.767808861901e+002 J/kgK
	density       : 1.033329304336e+003 kg/m^3
	heat capacity : 3.951778371490e+003 J/kgK	
	Ch. pot. H2O  : 1.448283288805e+004 J/kg
	therm. exp.   : 2.785224996784e-004 1/K
	compressib.   : 4.061297733553e-010 1/Pa
	lapse rate    : 1.999488253001e-008 K/Pa
	pot. temp.    : 1.961798732859e+001 degC
	*/

}





/************************************************************
Compute the partial derivative (d/dx)^nx (d/dy)^ny (d/dz)^nz
of the Sum gc(i,j,k) * x^i * y^j * z^k
at given point x, y, z
************************************************************/
double Gxyz(int nx, int ny, int nz, double x, double y, double z) 
{

	double gijk = 0;
	double xi ,yj ,zk, c;
	int imax, jmax, kmax, i, j, k, l;

	if (x == 0) imax = nx;
	else imax = maxS;
	if (y == 0) jmax = ny;
	else jmax = maxT;
	if (z == 0) kmax = nz;
	else kmax = maxP;


	xi = 1;
	for ( i = nx; i <= imax; ++i) {
		if ( i == 1) {
			if (i < imax) {
				xi = xi * x;
			}
			continue;
		}
			
		yj = 1;
		for (j = ny; j <= jmax; ++j) {

			zk = 1;
			for ( k = nz; k <= kmax; ++k) {

				if ( gc[i][j][k] ) {
					c = gc[i][j][k] * xi * yj * zk;
					for ( l = 1; l <= nx ; ++l) {
						c = c * (i - l + 1);
					}
					for ( l = 1; l <= ny ; ++l) {
						c = c * (j - l + 1);
					}
					for ( l = 1; l <= nz ; ++l) {
						c = c * (k - l + 1);
					}
					gijk = gijk + c;
				}

				if (k < kmax) {
					zk = zk * z;
				}
			}

			if (j < jmax) {
				yj = yj * y;
			}
		}

		if (i < imax) {
			xi = xi * x;
		}
	}

	//Gxyz = gijk;
	return (gijk);
}




/************************************************************
'nS, nT, nP  order of partial derivative in S, T, P
'S in psu, tabs in K, pabs in Pa
************************************************************/
double GSTP03(int nS, int nT, int nP, double S, double tabs, double pabs)
{
	double x, y, z;
	double gstp;
	double x2lnx;

	double Sref = 40;         //40 psu
	double Tref = 40;         //40C
	double Pref = 100000000;  //10 MPa


	x = sqrt(S / Sref);
	y = (tabs - 273.15) / Tref;
	z = (pabs - 101325) / Pref;


	//Derivatives of the polynomial part:
	switch(nS)		//consider only lowest salinity derivatives
	{
		case 0:
			gstp = Gxyz(0, nT, nP, x, y, z);
			break;
		case 1:
			if (x > 0) {
				gstp = Gxyz(1, nT, nP, x, y, z) / (2 * x);
			}
			else {
				gstp = Gxyz(2, nT, nP, x, y, z) / 2;
			}
			break;
		case 2:
			if (x > 0) {
				gstp = Gxyz(2, nT, nP, x, y, z) / pow(2 * x, 2);
				gstp = gstp - Gxyz(1, nT, nP, x, y, z) / (4 * x * x * x);
			}
			else {
				//2. S-Derivative at S = 0 Not Implemented
				printf("\n 2. S-Derivative at S = 0 Not Implemented \n");
			}
			break;
		default:
			//higher S-Derivatives Not Implemented
			printf("\n Higher S-Derivatives Not Implemented \n");
			break;
	}
     

	//Add derivatives of the logarithm term:
	//term with x^2*ln(x) is independent of pressure and linear in temperature)
	if ( (nP == 0) && (nT < 2) && (x > 0)) {

		x2lnx = 0;

		//S derivative
		switch (nS) 
		{
			case 0: 
				x2lnx = x * x * log(x);
				break;
			case 1:
				x2lnx = log(x) + 0.5;
				break;
			case 2:
				x2lnx = 1 / (2 * x * x);
				break;
			default:
				//' higher S-Derivatives Not Implemented
				printf("\n higher S-Derivatives Not Implemented \n");
				break;
		}

		//T derivative
		if (nT == 0) {
			gstp = gstp + (gc[1][0][0] + gc[1][1][0] * y) * x2lnx;
		}
		else {
			gstp = gstp + gc[1][1][0] * x2lnx;
		}
	}

	//GSTP03 = gstp / (Sref ^ nS * Tref ^ nT * Pref ^ nP)
	return ( gstp / ( pow(Sref,nS) * pow(Tref,nT) * pow(Pref,nP) ) );
}


/************************************************************
 Spsu in psu, tdegC in C, pdBar in dBar, prefdBar in dBar
************************************************************/
double PotTemp(double Spsu, double tdegC, double pdBar, double prefdBar)
{
	//=======================================================
	//compute potential temperatur (C) for reference pressure pref (dbar)
	//=======================================================

	double S;
	double T;
	double P;
	double Pref;
	double dt;
	double retval;
	double theta;
	double sigma;

	S = Spsu;
	T = tdegC + 273.15;
	P = pdBar * 10000 + 101325;
	Pref = prefdBar * 10000 + 101325;

	retval = -999;
	theta = T;
	sigma = -GSTP03(0, 1, 0, S, T, P);

	for (int iter = 1; iter <= 100; ++iter) {
		//Newton iteration
		dt = -(sigma + GSTP03(0, 1, 0, S, theta, Pref)) / GSTP03(0, 2, 0, S, theta, Pref);
		theta = theta + dt;
		if (fabs(dt) < 0.0001) {
			retval = theta - 273.15;
			break;
		}
	}
	return retval; 
}

/************************************************************
init coeffs: COEFFS03()
************************************************************/
void gc_init(void)
{
	for (int ii=0; ii<maxS+1;ii++) {
		for (int jj=0; jj<maxT+1;jj++) {
			for (int kk=0; kk<maxP+1;kk++) {
				gc[ii][jj][kk] = 0;
			}
		}
	}
	/*
         COEFFS03()
	 g(S,T,P) = (gc(1,0,0) + gc(1,1,0)*y)*x^2*ln(x) + Sum gc(i,j,k) * x^i * y^j * z^k
	                                                  ijk

	 S = 40 PSU * x^2, T = 40 C * y, P = 100 MPa * z
        */

	//IAPWS95:
	//CP(0,T,0)

	gc[0][2][0] = -12357.785933039;
	gc[0][3][0] = 736.741204151612;
	gc[0][4][0] = -148.185936433658;
	gc[0][5][0] = 58.0259125842571;
	gc[0][6][0] = -18.9843846514172;
	gc[0][7][0] = 3.05081646487967;

	//IAPWS95:
	//V(0,T,0)
	//V(0,T,P)
	//dV/dt (0,t,p)
	//dV/dt (0,t,0)
	//dV/dp (0,t,p)
	gc[0][0][1] = 100015.695367145;
	gc[0][1][1] = -270.983805184062;
	gc[0][2][1] = 1455.0364540468;
	gc[0][3][1] = -672.50778314507;
	gc[0][4][1] = 397.968445406972;
	gc[0][5][1] = -194.618310617595;
	gc[0][6][1] = 63.5113936641785;
	gc[0][7][1] = -9.63108119393062;
	gc[0][0][2] = -2544.5765420363;
	gc[0][1][2] = 776.153611613101;
	gc[0][2][2] = -756.558385769359;
	gc[0][3][2] = 499.360390819152;
	gc[0][4][2] = -301.815380621876;
	gc[0][5][2] = 120.520654902025;
	gc[0][6][2] = -22.2897317140459;
	gc[0][0][3] = 284.517778446287;
	gc[0][1][3] = -196.51255088122;
	gc[0][2][3] = 273.479662323528;
	gc[0][3][3] = -239.545330654412;
	gc[0][4][3] = 152.196371733841;
	gc[0][5][3] = -55.2723052340152;
	gc[0][6][3] = 8.17060541818112;
	gc[0][0][4] = -33.3146754253611;
	gc[0][1][4] = 28.9796526294175;
	gc[0][2][4] = -55.5604063817218;
	gc[0][3][4] = 48.8012518593872;
	gc[0][4][4] = -26.3748377232802;
	gc[0][5][4] = 6.48190668077221;
	gc[0][0][5] = 4.20263108803084;
	gc[0][1][5] = -2.13290083518327;
	gc[0][2][5] = 4.34420671917197;
	gc[0][3][5] = -1.66307106208905;
	gc[0][0][6] = -0.546428511471039;

	//IAPWS95 Reference State: energy=0 and entropy=0 at triple point of liquid water:
	gc[0][0][0] = 101.342743139672;
	gc[0][1][0] = 5.90578348518236;

	//enthalpy=0 and entropy=0 standard seawater
	//g(35,0,0) = 0:
	gc[2][0][0] = 1376.28030233939;
	//dg(35,0,0)/dt = 0:
	gc[2][1][0] = 140.576997717291;

	//MPD73 CP(S,T,0)
	//P80 V(S,T,0)
	//M76 V(S,T,0)
	//PG93 V(S,T,0)
	//Table IV
	//Table I-III
	//Table V-VI
	//Caldwell TMD
	//CM76 V(S,T,P)
	//BS70 V(S,t,P) - V(S,0,P)
	//BDSW70 CP(S,t,0)
	gc[2][2][0] = 929.460016974089;
	gc[3][2][0] = -158.720177628421;
	gc[2][3][0] = -260.427286048143;
	gc[3][3][0] = 67.5232147262047;
	gc[2][4][0] = 97.1562727658403;
	gc[3][4][0] = -16.8901274896506;
	gc[2][5][0] = -17.43743842213;
	gc[2][0][1] = -3310.49154044839;
	gc[3][0][1] = 199.459603073901;
	gc[4][0][1] = -54.7919133532887;
	gc[5][0][1] = 36.0284195611086;
	gc[2][1][1] = 729.116529735046;
	gc[3][1][1] = -175.292041186547;
	gc[4][1][1] = -22.6683558512829;
	gc[2][2][1] = -860.764303783977;
	gc[3][2][1] = 383.058066002476;
	gc[2][3][1] = 694.244814133268;
	gc[3][3][1] = -460.319931801257;
	gc[2][4][1] = -297.728741987187;
	gc[3][4][1] = 234.565187611355;
	gc[2][0][2] = 384.794152978599;
	gc[3][0][2] = -52.2940909281335;
	gc[4][0][2] = -4.08193978912261;
	gc[2][1][2] = -343.956902961561;
	gc[3][1][2] = 83.1923927801819;
	gc[2][2][2] = 337.409530269367;
	gc[3][2][2] = -54.1917262517112;
	gc[2][3][2] = -204.889641964903;
	gc[2][4][2] = 74.726141138756;
	gc[2][0][3] = -96.5324320107458;
	gc[3][0][3] = 68.0444942726459;
	gc[4][0][3] = -30.1755111971161;
	gc[2][1][3] = 124.687671116248;
	gc[3][1][3] = -29.483064349429;
	gc[2][2][3] = -178.314556207638;
	gc[3][2][3] = 25.6398487389914;
	gc[2][3][3] = 113.561697840594;
	gc[2][4][3] = -36.4872919001588;
	gc[2][0][4] = 15.8408172766824;
	gc[3][0][4] = -3.41251932441282;
	gc[2][1][4] = -31.656964386073;
	gc[2][2][4] = 44.2040358308;
	gc[2][3][4] = -11.1282734326413;
	gc[2][0][5] = -2.62480156590992;
	gc[2][1][5] = 7.04658803315449;
	gc[2][2][5] = -7.92001547211682;

	//limiting laws:
	gc[1][0][0] = 5813.28667992895;
	gc[1][1][0] = 851.295871122672;
	gc[3][0][0] = -2432.0947227047;
	gc[3][1][0] = -493.512590658728;

	//Colligative Properties
	//Fit of Freezing Points DK74
	//Fit of Dilution Heat Bromley
	//Fit of Dilution Heat MHH
	gc[4][0][0] = 2630.93863474177;
	gc[5][0][0] = -2559.89065469719;
	gc[6][0][0] = 1695.91780114244;
	gc[7][0][0] = -466.680815621115;
	gc[4][1][0] = 845.15825213234;
	gc[5][1][0] = -810.552561548477;
	gc[6][1][0] = 506.103588839417;
	gc[7][1][0] = -129.049444012372;

}


