Attribute VB_Name = "Gibbs_03"
'#####################################################################
'#                                                                   #
'#    written by Rainer Feistel, 03 Nov 2003                         #
'#                                                                   #
'#    VB 6.0 demo program to compute thermodynamic properties        #
'#    of seawater from a thermodynamic potential function g(S,t,p).  #
'#    Details are published in "Progress in Oceanography",           #
'#    58 (2003) 43-114, entitled                                     #
'#                                                                   #
'#    "A new extended Gibbs thermodynamic potential of seawater"     #
'#                                                                   #
'#               by R.Feistel                                        #
'#                                                                   #
'#                                                                   #
'#                 Dr. Rainer Feistel                                #
'#                 Institut fuer Ostseeforschung, IOW                #
'#                 D-18119 Warnemuende                               #
'#                 Seestrasse 15                                     #
'#                 GERMANY                                           #
'#                                                                   #
'#                 Tel: +49-381-5197-152                             #
'#                 Fax: +49-381-5197-4818                            #
'#                 Email: rainer.feistel@io-warnemuende.de           #
'#                                                                   #
'#####################################################################

'Fortran convention for 16 Bit INTEGER
DefInt I-N
'All other variables are DOUBLE PRECISION by default
DefDbl A-H, O-Z

'Specify a COMMON block for array of polynomial coefficients gc(7,7,6)
Public Const maxS = 7
Public Const maxT = 7
Public Const maxP = 6
Public gc(maxS, maxT, maxP)

'error flag
Public Const ErrorReturn = -9.999999999999E+99

Function Gxyz(ByVal nx, ByVal ny, ByVal nz, ByVal x, ByVal y, ByVal z) As Double
'============================================================
'Compute the partial derivative (d/dx)^nx (d/dy)^ny (d/dz)^nz
'of the Sum gc(i,j,k) * x^i * y^j * z^k
'at given point x, y, z
'============================================================

gijk = 0
If x = 0 Then imax = nx Else imax = maxS
If y = 0 Then jmax = ny Else jmax = maxT
If z = 0 Then kmax = nz Else kmax = maxP
xi = 1
For i = nx To imax
  'skip coefficients c(1,J,K), they belong to xlnx , treated separately
  If i = 1 Then GoTo nexti

  yj = 1
  For J = ny To jmax

    zk = 1
    For K = nz To kmax

      If gc(i, J, K) <> 0 Then
        c = gc(i, J, K) * xi * yj * zk
        For l = 1 To nx
          c = c * (i - l + 1)
        Next l
        For l = 1 To ny
          c = c * (J - l + 1)
        Next l
        For l = 1 To nz
          c = c * (K - l + 1)
        Next l
        gijk = gijk + c
      End If

      If K < kmax Then zk = zk * z
    Next K

    If J < jmax Then yj = yj * y
  Next J

nexti:
  If i < imax Then xi = xi * x
Next i

Gxyz = gijk
End Function

Function GSTP03(ByVal nS, ByVal nT, ByVal nP, ByVal S, ByVal tabs, ByVal pabs) As Double
'======================================================
'computes partial derivatives of specific free enthalpy
'nS, nT, nP  order of partial derivative in S, T, P
'S in psu, tabs in K, pabs in Pa
'======================================================

Const Sref = 40         '40 psu
Const Tref = 40         '40C
Const Pref = 100000000  '100 MPa

GSTP03 = ErrorReturn  'preset: flag for error return

'exclude crudely erratic calls:
If nS < 0 Or nT < 0 Or nP < 0 Then Exit Function
If S < 0 Or tabs <= 0 Or pabs < 0 Then Exit Function
If S = 0 And nS > 0 And nP = 0 And nT < 2 Then Exit Function

x = Sqr(S / Sref)
y = (tabs - 273.15) / Tref
z = (pabs - 101325) / Pref

'Derivatives of the polynomial part:
Select Case nS  'consider only lowest salinity derivatives

  Case 0:
          gstp = Gxyz(0, nT, nP, x, y, z)

  Case 1:
          If x > 0 Then
            gstp = Gxyz(1, nT, nP, x, y, z) / (2 * x)
          Else
            gstp = Gxyz(2, nT, nP, x, y, z) / 2
          End If
  
  Case 2:
          If x > 0 Then
            gstp = Gxyz(2, nT, nP, x, y, z) / (2 * x) ^ 2
            gstp = gstp - Gxyz(1, nT, nP, x, y, z) / (4 * x ^ 3)
          Else
            ' 2. S-Derivative at S = 0 Not Implemented
            Exit Function
          End If

  Case Else: ' higher S-Derivatives Not Implemented
          Exit Function

End Select

'Add derivatives of the logarithm term:
'(term with x^2*ln(x) is independent of pressure and linear in temperature)
If nP = 0 And nT < 2 And x > 0 Then

  x2lnx = 0
  'S derivative
  Select Case nS  'consider only lowest salinity derivatives
    Case 0: x2lnx = x ^ 2 * Log(x)
    Case 1: x2lnx = Log(x) + 0.5
    Case 2: x2lnx = 1 / (2 * x ^ 2)
    Case Else: ' higher S-Derivatives Not Implemented
            Exit Function
  End Select

  'T derivative
  If nT = 0 Then
    gstp = gstp + (gc(1, 0, 0) + gc(1, 1, 0) * y) * x2lnx
  Else
    gstp = gstp + gc(1, 1, 0) * x2lnx
  End If

End If

GSTP03 = gstp / (Sref ^ nS * Tref ^ nT * Pref ^ nP)

End Function


Sub COEFFS03()
Rem g(S,T,P) = (gc(1,0,0) + gc(1,1,0)*y)*x^2*ln(x) + Sum gc(i,j,k) * x^i * y^j * z^k
Rem                                                  ijk
Rem
Rem S = 40 PSU * x^2, T = 40 C * y, P = 100 MPa * z

'IAPWS95:
'CP(0,T,0)
gc(0, 2, 0) = -12357.785933039
gc(0, 3, 0) = 736.741204151612
gc(0, 4, 0) = -148.185936433658
gc(0, 5, 0) = 58.0259125842571
gc(0, 6, 0) = -18.9843846514172
gc(0, 7, 0) = 3.05081646487967

'IAPWS95:
'V(0,T,0)
'V(0,T,P)
'dV/dt (0,t,p)
'dV/dt (0,t,0)
'dV/dp (0,t,p)
gc(0, 0, 1) = 100015.695367145
gc(0, 1, 1) = -270.983805184062
gc(0, 2, 1) = 1455.0364540468
gc(0, 3, 1) = -672.50778314507
gc(0, 4, 1) = 397.968445406972
gc(0, 5, 1) = -194.618310617595
gc(0, 6, 1) = 63.5113936641785
gc(0, 7, 1) = -9.63108119393062
gc(0, 0, 2) = -2544.5765420363
gc(0, 1, 2) = 776.153611613101
gc(0, 2, 2) = -756.558385769359
gc(0, 3, 2) = 499.360390819152
gc(0, 4, 2) = -301.815380621876
gc(0, 5, 2) = 120.520654902025
gc(0, 6, 2) = -22.2897317140459
gc(0, 0, 3) = 284.517778446287
gc(0, 1, 3) = -196.51255088122
gc(0, 2, 3) = 273.479662323528
gc(0, 3, 3) = -239.545330654412
gc(0, 4, 3) = 152.196371733841
gc(0, 5, 3) = -55.2723052340152
gc(0, 6, 3) = 8.17060541818112
gc(0, 0, 4) = -33.3146754253611
gc(0, 1, 4) = 28.9796526294175
gc(0, 2, 4) = -55.5604063817218
gc(0, 3, 4) = 48.8012518593872
gc(0, 4, 4) = -26.3748377232802
gc(0, 5, 4) = 6.48190668077221
gc(0, 0, 5) = 4.20263108803084
gc(0, 1, 5) = -2.13290083518327
gc(0, 2, 5) = 4.34420671917197
gc(0, 3, 5) = -1.66307106208905
gc(0, 0, 6) = -0.546428511471039
gc(0, 0, 1) = 100015.695367145
gc(0, 1, 1) = -270.983805184062
gc(0, 2, 1) = 1455.0364540468
gc(0, 3, 1) = -672.50778314507
gc(0, 4, 1) = 397.968445406972
gc(0, 5, 1) = -194.618310617595
gc(0, 6, 1) = 63.5113936641785
gc(0, 7, 1) = -9.63108119393062
gc(0, 0, 2) = -2544.5765420363
gc(0, 1, 2) = 776.153611613101
gc(0, 2, 2) = -756.558385769359
gc(0, 3, 2) = 499.360390819152
gc(0, 4, 2) = -301.815380621876
gc(0, 5, 2) = 120.520654902025
gc(0, 6, 2) = -22.2897317140459
gc(0, 0, 3) = 284.517778446287
gc(0, 1, 3) = -196.51255088122
gc(0, 2, 3) = 273.479662323528
gc(0, 3, 3) = -239.545330654412
gc(0, 4, 3) = 152.196371733841
gc(0, 5, 3) = -55.2723052340152
gc(0, 6, 3) = 8.17060541818112
gc(0, 0, 4) = -33.3146754253611
gc(0, 1, 4) = 28.9796526294175
gc(0, 2, 4) = -55.5604063817218
gc(0, 3, 4) = 48.8012518593872
gc(0, 4, 4) = -26.3748377232802
gc(0, 5, 4) = 6.48190668077221
gc(0, 0, 5) = 4.20263108803084
gc(0, 1, 5) = -2.13290083518327
gc(0, 2, 5) = 4.34420671917197
gc(0, 3, 5) = -1.66307106208905
gc(0, 0, 6) = -0.546428511471039

'IAPWS95 Reference State: energy=0 and entropy=0 at triple point of liquid water:
gc(0, 0, 0) = 101.342743139672
gc(0, 1, 0) = 5.90578348518236

'enthalpy=0 and entropy=0 standard seawater
'g(35,0,0) = 0:
gc(2, 0, 0) = 1376.28030233939
'dg(35,0,0)/dt = 0:
gc(2, 1, 0) = 140.576997717291

'MPD73 CP(S,T,0)
'P80 V(S,T,0)
'M76 V(S,T,0)
'PG93 V(S,T,0)
'Table IV
'Table I-III
'Table V-VI
'Caldwell TMD
'CM76 V(S,T,P)
'BS70 V(S,t,P) - V(S,0,P)
'BDSW70 CP(S,t,0)
gc(2, 2, 0) = 929.460016974089
gc(3, 2, 0) = -158.720177628421
gc(2, 3, 0) = -260.427286048143
gc(3, 3, 0) = 67.5232147262047
gc(2, 4, 0) = 97.1562727658403
gc(3, 4, 0) = -16.8901274896506
gc(2, 5, 0) = -17.43743842213
gc(2, 0, 1) = -3310.49154044839
gc(3, 0, 1) = 199.459603073901
gc(4, 0, 1) = -54.7919133532887
gc(5, 0, 1) = 36.0284195611086
gc(2, 1, 1) = 729.116529735046
gc(3, 1, 1) = -175.292041186547
gc(4, 1, 1) = -22.6683558512829
gc(2, 2, 1) = -860.764303783977
gc(3, 2, 1) = 383.058066002476
gc(2, 3, 1) = 694.244814133268
gc(3, 3, 1) = -460.319931801257
gc(2, 4, 1) = -297.728741987187
gc(3, 4, 1) = 234.565187611355
gc(2, 0, 2) = 384.794152978599
gc(3, 0, 2) = -52.2940909281335
gc(4, 0, 2) = -4.08193978912261
gc(2, 1, 2) = -343.956902961561
gc(3, 1, 2) = 83.1923927801819
gc(2, 2, 2) = 337.409530269367
gc(3, 2, 2) = -54.1917262517112
gc(2, 3, 2) = -204.889641964903
gc(2, 4, 2) = 74.726141138756
gc(2, 0, 3) = -96.5324320107458
gc(3, 0, 3) = 68.0444942726459
gc(4, 0, 3) = -30.1755111971161
gc(2, 1, 3) = 124.687671116248
gc(3, 1, 3) = -29.483064349429
gc(2, 2, 3) = -178.314556207638
gc(3, 2, 3) = 25.6398487389914
gc(2, 3, 3) = 113.561697840594
gc(2, 4, 3) = -36.4872919001588
gc(2, 0, 4) = 15.8408172766824
gc(3, 0, 4) = -3.41251932441282
gc(2, 1, 4) = -31.656964386073
gc(2, 2, 4) = 44.2040358308
gc(2, 3, 4) = -11.1282734326413
gc(2, 0, 5) = -2.62480156590992
gc(2, 1, 5) = 7.04658803315449
gc(2, 2, 5) = -7.92001547211682

'limiting laws:
gc(1, 0, 0) = 5813.28667992895
gc(1, 1, 0) = 851.295871122672
gc(3, 0, 0) = -2432.0947227047
gc(3, 1, 0) = -493.512590658728

'Colligative Properties
'Fit of Freezing Points DK74
'Fit of Dilution Heat Bromley
'Fit of Dilution Heat MHH
gc(4, 0, 0) = 2630.93863474177
gc(5, 0, 0) = -2559.89065469719
gc(6, 0, 0) = 1695.91780114244
gc(7, 0, 0) = -466.680815621115
gc(4, 1, 0) = 845.15825213234
gc(5, 1, 0) = -810.552561548477
gc(6, 1, 0) = 506.103588839417
gc(7, 1, 0) = -129.049444012372

End Sub

Sub F03demo(ByVal S_psu As Double, _
            ByVal t_degC As Double, _
            ByVal p_dbar As Double)

Dim S As Double, T As Double, P As Double

Call COEFFS03

S = S_psu                    'psu  -> psu
T = t_degC + 273.15          'degC -> K
P = p_dbar * 10000# + 101325# 'dBar -> Pa

Debug.Print "S=", S_psu; "psu"
Debug.Print "T=", t_degC; "C"
Debug.Print "P=", p_dbar; "dbar"
Debug.Print ""
Debug.Print "free enthalpy", GSTP03(0, 0, 0, S, T, P); "J/kg"
Debug.Print "chem. pot.   ", GSTP03(1, 0, 0, S, T, P); "J/kg psu"
Debug.Print "entropy      ", -GSTP03(0, 1, 0, S, T, P); "J/kgK"
Debug.Print "density      ", 1# / GSTP03(0, 0, 1, S, T, P); "kg/m^3"
Debug.Print "heat capacity", -T * GSTP03(0, 2, 0, S, T, P); "J/kgK"
Debug.Print "Ch. pot. H2O ", GSTP03(0, 0, 0, S, T, P) - S * GSTP03(1, 0, 0, S, T, P); "J/kg"
Debug.Print "therm. exp.  ", GSTP03(0, 1, 1, S, T, P) / GSTP03(0, 0, 1, S, T, P); "1/K"
Debug.Print "compressib.  ", -GSTP03(0, 0, 2, S, T, P) / GSTP03(0, 0, 1, S, T, P); "1/Pa"
Debug.Print "lapse rate   ", -GSTP03(0, 1, 1, S, T, P) / GSTP03(0, 2, 0, S, T, P); "K/Pa"
Debug.Print "pot. temp.   ", PotTemp(S_psu, t_degC, p_dbar); "C"

'Printout example:
'S=             35 psu
'T=             20 C
'P=             2000 dbar
'
'free enthalpy  16583.1806714797 J/kg
'chem. pot.     60.0099366692805 J/kg psu
'entropy        276.780886190056 J/kgK
'density        1033.32930433584 kg/m^3
'heat capacity  3951.77837149032 J/kgK
'Ch. pot. H2O   14482.8328880549 J/kg
'therm. exp.    2.78522499678412E-04 1/K
'compressib.    4.06129773355324E-10 1/Pa
'lapse rate     1.99948825300137E-08 K/Pa
'pot. temp.     19.617987328589 C

End Sub
Function PotTemp(ByVal S_psu, ByVal t_degC, ByVal p_dbar, Optional ByVal pref_dbar = 0)
'=======================================================
'compute potential temperatur (C) for reference pressure pref (dbar)
'=======================================================

If gc(0, 0, 0) = 0 Then Call COEFFS03

S = S_psu
T = t_degC + 273.15
P = p_dbar * 10000# + 101325#
Pref = pref_dbar * 10000# + 101325#

PotTemp = ErrorReturn

theta = T
sigma = -GSTP03(0, 1, 0, S, T, P)

For iter = 1 To 100
  dt = -(sigma + GSTP03(0, 1, 0, S, theta, Pref)) / GSTP03(0, 2, 0, S, theta, Pref) 'Newton iteration
  theta = theta + dt
  If Abs(dt) < 0.0001 Then
    PotTemp = theta - 273.15
    Exit Function
  End If
Next iter

End Function


